package org.pnyka.cryptolib.ZKP;

import java.math.BigInteger;
import org.bouncycastle.crypto.params.ElGamalParameters;
import org.bouncycastle.crypto.params.ElGamalPrivateKeyParameters;
import org.bouncycastle.crypto.params.ElGamalPublicKeyParameters;


/**
 * 
 * Proof o ElGamal encryption of a single bit and of a number in an interval.
 * 
 * Prover knows the 2-tuple (g<sup>a</sup>, h<sup>a</sup>i<sup>1</sup>)
 * (and in fact knows the value of every letter g, h, i,  in it) 
 * which is an Elgamal encryption of the message i<sup>1</sup>.
 * Here g, h, i are public group elements in a public elliptic
 * curve group of large prime order G and  is a random integer mod G.
 * Note this is the Elgamal encryption of a single bit (the  sign).
 * Peter wishes to prove to Verifier that this 2-tuple (s, t) is the 
 * Elgamal encryption of either i or i<sup>-1</sup> but without revealing
 * which, and without revealing the random (mod G) integer . This comes down 
 * to proving that log<sub>g</sub> s = log<sub>h</sub>(t/i) or =
 * log<sub>h</sub>(ti).
 * 
 * Here is the proof protocol:<ol>
 * <li>1. Prover announces the 2-tuple (s, t) = (g<sup>a</sup>, h<sup>a</sup>i<sup>1</sup>).</li>
 * <li>2. Prover chooses w, r, d at random mod G</li>
 * <li>3. Prover computes a1 = g<sup>r</sup>s<sup>d</sup>; b1 = h<sup>r</sup>(ti<sup>1</sup>)<sup>d</sup></li>
 * <li>4. Prover computes a2 <- g<sup>w</sup>; b2 <- h<sup>w</sup></li>
 * <li>5. If -1 then Prover swaps (a<sub>1</sub>, b<sub>1</sub>) <-> (a<sub>2</sub>, b<sub>2</sub>)</li>
 * <li>6. Prover sends (a<sub>1</sub>, b<sub>1</sub>, a<sub>2</sub>, b<sub>2</sub>) to Verifier</li>
 * <li>7. Verifier selects a challenge integer c mod G at random and sends it to Prover</li>
 * <li>8. Prover computes e = c - d; q = w - e</li>
 * <li>9. If -1 then Peter swaps (d, r) <-> (e, q)</li>
 * <li>10. Prover sends (d, e, r, q) to Verifier</li>
 * <li>11. Vera verifies c = d + e, a<sub>1</sub> = g<sup>r</sup>s<sup>d</sup>, b<sub>1</sub> =  h<sup>r</sup>(ti<sup>d</sup>)<sup>d</sup>,
 * <li>a<sub>2</sub> = g<sup>q</sup>s<sup>e</sup>, and b<sub>2</sub> = h<sup>q</sup>(t/i)<sup>e</sup>. </li>
 * <ol>
 */
public class BitInterZKPAlg extends BaseZKP
{
    private static final BigInteger ZERO = BigInteger.valueOf(0);
    private static final BigInteger ONE  = BigInteger.valueOf(1);
    private static final BigInteger TWO  = BigInteger.valueOf(2);
   
    public BigInteger s, t, a1, a2, b1, b2, c, d, e, r, q;
    private boolean verification;
   
    /**
    * @param p parameter p
    * @param g parameter g
    * @param h parameter h
    */   
    public BitInterZKPAlg(BigInteger p, BigInteger g, BigInteger h)
    {
        super(p, g, h);
    }
   
   /**
    * Step 1 of ZKP.
    * @param a random number
    * @param i value encrypted with ElGamal algorithm
    * @param bit 1 or -1
    */
   public void computeST(BigInteger a, BigInteger i, int bit)
   {
       s = g.modPow(p, a);//ypologismos tou s
       if(bit == 1)
            t = h.modPow(a, p).multiply(i).mod(p);
       else if(bit == -1)
            t = i.modInverse(p).multiply(h.modPow(a, p)).mod(p);
       
   }
   
   
   /**
    * Steps 3 and 4 of ZKP.
    * @param w a random value mod G
    * @param r a random value mod G
    * @param d a random value mod G
    * @param i value encrypted with ElGamal algorithm
    * @param bit 1 or -1
    */
   
   public void computeAB(BigInteger w, BigInteger r, BigInteger d, BigInteger i, int bit)
   {
       this.r = r;
       this.d = d;

       a1 = g.modPow(r, p).multiply(s.modPow(d, p)).mod(p);
       a2 = g.modPow(w, p);
       
       if(bit == 1)
           b1 = h.modPow(r, p).multiply((t.multiply(i).mod(p)).modPow(d, p)).mod(p);
       else if(bit == -1)
           b1 = h.modPow(r, p).multiply(t.multiply(i.modInverse(p)).modPow(d, p)).mod(p);
       
       b2 = h.modPow(w, p);
       
       if(bit == -1)
       {
           BigInteger temp1, temp2;
           
           temp1 = a1;
           a1 = a2;
           a2 = temp1;
           
           temp2 = b1;
           b1 = b2;
           b2 = temp2;
       }
               
          
   }
   
   /**
     * Step 3 of ZKP.
     */
    public void setChallenge()
    {
        c = BaseZKP.selectRandom(p);
    }
    
   /**
    * Step 8 of ZKP.
    * @param c the challenge sent by verifier
    * @param wrd the array with random w, r, d values
    * @param a the secret value that prover chooses
    * @return a 2-field arraywith e, q values
    *
    */
   public void computeDERQ(BigInteger w, BigInteger a, int bit)
   {
       e = c.subtract(d).mod(p);
       q = w.subtract(a.multiply(e));
       
       if(bit == -1)
       {
           BigInteger temp1, temp2;
           
           temp1 = d;
           d = e;
           e = temp1;
           
           temp2 = r;
           r = q;
           q = temp2; 
       }
      
   }
   
   /**
    * Does the verification of the values sent by the prover. Step 11 of ZKP
    * @param i value encrypted with ElGamal algorithm
    * @return verification result
    */
   public boolean verify(BigInteger i)
   {
       BigInteger cVer, a1Ver, a2Ver = null, b1Ver, b2Ver;
              
       cVer  = d.add(e);
       a1Ver = g.modPow(r, p).multiply(s.modPow(d, p)).mod(p);
       b1Ver = h.modPow(r, p).multiply((t.multiply(i)).modPow(d,p)).mod(p);       
       
       a2Ver = g.modPow(q, p).multiply(s.modPow(e, p)).mod(p);
       b2Ver = h.modPow(q, p).multiply((t.multiply(i.modInverse(p))).modPow(e, p)).mod(p);
          
       if(c.equals(cVer) && a1.equals(a1Ver) && a2.equals(a2Ver) && b1.equals(b1Ver) && b2.equals(b2Ver))
           verification = true;
       else 
           verification = false;
       
       return verification;
   }
         
    
}