package org.pnyka.cryptolib.prng;

import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Random;

/**
 * The Blum-Blum-Shub random number generator.
 *
 * <p>
 * The Blum-Blum-Shub is a "cryptographically secure" random number
 * generator.
 * </p>
 *
 * <p>
 * The Algorithm:
 * </p>
 * <ol>
 * <li>
 *  (setup) generate two secret prime numbers <i>p</i>, <i>q</i> such that
 *   <i>p</i> &ne; <i>q</i>, <i>p</i> &equiv; 3 mod 4, <i>q</i> &equiv; 3 mod 4.
 * </li>
 * <li> (setup) compute <i>n</i> = <i>pq</i>.  <i>n</i> can be re-used, but
 *   <i>p</i>, and <i>q</i> are secret and should be disposed of.</li>
 * <li> Generate a (secure) random seed <i>s</i> in the range [1, <i>n</i> -1]
 *   such that gcd(<i>s</i>, <i>n</i>) = 1.
 * <li> Compute <i>x</i> = <i>s</i><sup>2</sup> mod <i>n</i></li>
 * <li> Compute a single random bit with:
 *   <ol>
 *   <li> <i>x</i> = <i>x</i><sup>2</sup> mod <i>n</i></li>
 *   <li> return Least-Significant-Bit(<i>x</i>) (i.e. <i>x</i> & 1)</li>
 *   </ol>
 * Repeat as necessary.
 * </li>
 * </ol>
 *
 * <p>
 * More details are in  the <a href="http://www.cacr.math.uwaterloo.ca/hac/"><i>Handbook of Applied Cryptography</i></a>,
 * <a href="http://www.cacr.math.uwaterloo.ca/hac/about/chap5.pdf">Section 5.5.2</a>
 * </p>
 *

 *
 */

public class BBSRandomGenerator implements RandomGenerator
{
    // pre-compute a few values
    private static final BigInteger two = BigInteger.valueOf(2L);
    private static final BigInteger three = BigInteger.valueOf(3L);
    private static final BigInteger four = BigInteger.valueOf(4L);

    /**
     * main parameters
     */
    private BigInteger n;
    private BigInteger state;
             
 
    /**
     * Constructor, generates prime and seed
     *
     * @param bits
     */
    public BBSRandomGenerator(int bits) //Constructor 1
    {
	this(bits, new Random());
    }

    /**
     * Constructor, generates prime and seed
     *
     * @param bits
     * @param rand
     */
    public BBSRandomGenerator(int bits, Random rand)  //Constructor 2
    {
	this(generateN(bits, rand));
    }

    /**
     * A constructor to specify the "n-value" to the Blum-Blum-Shub algorithm.
     * The initial seed is computed using Java's internal "true" random number
     * generator.
     *
     * @param n
     *            The n-value.
     */
    public BBSRandomGenerator(BigInteger n)  //Constructor 3
    {
	this(n, SecureRandom.getSeed(n.bitLength() / 8));
    }

    /**
     * A constructor to specify both the n-value and the seed to the
     * Blum-Blum-Shub algorithm. 
     *
     * @param n The n-value using a BigInteger
     * @param seed The seed value using a byte[] array.
     */
    public BBSRandomGenerator(BigInteger n, byte[] seed)  //Constructor 4
    {
	this.n = n;
	setSeed(seed);
    }
        
   
    /**
     * Sets or resets generator's internal state
     *
     * @param seedBytes The new seed.
     */
    public void setSeed(byte[] seedBytes) //seedBytes: a byte array which contains the seed
    {
	BigInteger seed = new BigInteger(1, seedBytes);
	state = seed.mod(n);
    }
    
    
    /**
     * Generates appropriate prime number for use in Blum-Blum-Shub.
     *
     * This generates the appropriate primes (p = 3 mod 4) needed to compute the
     * "n-value" for Blum-Blum-Shub.
     *
     * @param bits Number of bits in prime
     * @param rand A source of randomness
     */
    private static BigInteger getPrime(int bits, Random rand) 
    {
	BigInteger p;
	while (true) 
        {
	    p = new BigInteger(bits, 100, rand);
	    if (p.mod(four).equals(three))
		break;
	}
	return p;
    }

    /**
     * Generates the "n value" -- the multiplication of two equally sized
     * random prime numbers -- for use in the Blum-Blum-Shub algorithm.
     *
     * @param bits The number of bits of security
     * @param rand A random instance to aid in generating primes
     * @return A BigInteger, the <i>n</i>.
     */
    public static BigInteger generateN(int bits, Random rand) 
    {
	BigInteger p = getPrime(bits/2, rand);
	BigInteger q = getPrime(bits/2, rand);

	// make sure p != q (almost always true, but just in case, check)
	while (p.equals(q)) 
        {
	    q = getPrime(bits, rand);
	}
        
	return p.multiply(q);
    }
   
    
     /**
     * Returns a random number 
     *
     * See also java.util.Random#next
     * @param numBits the length in bits of random number
     * @return a random Biginteger
     */
    public BigInteger nextNumber(int numBits)
    {
        BigInteger result = BigInteger.ZERO;
        int result2 = 0;
        
	for (int i = numBits; i != 0; --i) 
        {
	    state = state.modPow(two, n);
	    result = result.shiftLeft(1).or((state.testBit(0) == true ? BigInteger.ONE : BigInteger.ZERO));
            result2 = (result2 << 1) | (state.testBit(0) == true ? 1 : 0);
	}
        
        return result;
    }

   
}

