package org.pnyka.cryptolib.Signatures;

import java.math.BigInteger;

import java.security.SecureRandom;

import org.bouncycastle.crypto.CipherParameters;
import org.bouncycastle.crypto.params.ElGamalKeyParameters;
import org.bouncycastle.crypto.params.ElGamalPrivateKeyParameters;
import org.bouncycastle.crypto.params.ElGamalPublicKeyParameters;
import org.bouncycastle.crypto.params.ParametersWithRandom;
import org.pnyka.cryptolib.crypto.CipherText;

/**
 * ElGamal Digital Signature Scheme.
 *
 * <ol>
 *  The algorithm consists of the following three steps: 
 *  <li>The generation of Elgamal parameters p, g, of private and public key. These parameters have the same properties to the ElGamal encryption algorithm parameters.</li>
 *  <li>Signature generation. To sign a message m the signer performs the following steps.
 *      <ol>
 *          <li>Choose a random <i>k</i> such that 0 &lt; <i>k</i> &lt; <i>p</i>&#8722; 1 and <i>gcd</i>(<i>k</i>,<i>p</i> &#8722; 1) = 1</li>
 *          <li>Choose a random <i>k</i> such that 0 &lt; <i>k</i> &lt; <i>p</i> &#8722; 1 and <i>gcd</i>(<i>k</i>,<i>p</i> &#8722; 1) = 1</li>
 *          <li>Compute: r = <i>g<sup>k</sup></i> mod <i>p.</i></li>
 *          <li>Compute:<i> s</i> = (<i>H</i>(<i>m</i>)  <i>x</i>*<i>r</i>)*<i>k</i><sup>-1</sup>&nbsp;(mod p-1)</li>
 *          <li>If <i>s</i> = 0 start over again.</li>
 *      </ol>
 *      <p>Then the pair <i>(r,s)</i> is the digital signature of <i>m</i>. The signer repeats these steps for every signature.</p>
 *  </li>
 *  <li>Signature verification. A signature <i>(r,s)</i> of a message <i>m</i> is verified as follows.
 *      <ol>
 *          <li>0 &lt; <i>r</i> &lt; <i>p</i> &#954;&#945;&#953;&nbsp; 0 &lt; <i>s</i> &lt; <i>p</i> &#8722; 1</li>
 *          <li><i>g</i><sup>H(m) </sup>= <i>y</i><sup>r </sup>* <i>r</i><sup>s&nbsp; </sup>(mod<i>p</i>)</li>
 *      </ol>
 *      <p>The verifier accepts a signature if all conditions are satisfied and rejects it otherwise.</p>
 *  </li>
 *  <p>In this implementation instead of using the hash value of a message m (H(m)) we use the encrypted value of the message m.</p>
 * </ol>
 * 
 */

public class ElGamalSigner
{
    private ElGamalKeyParameters key;
    private SecureRandom    random;
    
     /**
     * Initializes the signer for signature generation or signature
     * verification.
     *
     * @param forSigning true if we are generating a signature, false otherwise.
     * @param param key parameters for signature generation.
     */
    public void init(boolean forSigning, CipherParameters param) 
    {
        if (forSigning)
        {
            if (param instanceof ParametersWithRandom)
            {
                ParametersWithRandom    rParam = (ParametersWithRandom)param;

                this.random = rParam.getRandom();
                this.key = (ElGamalPrivateKeyParameters)rParam.getParameters();
            }
            else
            {
                this.random = new SecureRandom();
                this.key = (ElGamalPrivateKeyParameters)param;
            }
        }
        else
        {    
            this.key = (ElGamalPublicKeyParameters)param;
        }
    }
    
    /**
     * Signs the passed in message 
     *
     * @param cipher the cipher to be signed.
     * @return a two-field array of big integers representing the r and s values respectively.
     */
    public BigInteger[] generateSignature(CipherText cipher) //message = M''
    {    
        BigInteger      c = cipher.getC();
        
        BigInteger      p = key.getParameters().getP();
        BigInteger      g = key.getParameters().getG();
        
        BigInteger      k;
        
        int pBitLength = p.bitLength();
        k = new BigInteger(pBitLength, random);
       
        while (k.equals(BigInteger.ZERO) || (k.compareTo(p.subtract(BigInteger.valueOf(1))) > 0) || k.gcd(p.subtract(BigInteger.ONE)).compareTo(BigInteger.ONE)!=0)
        {
             k = new BigInteger(pBitLength, random);
        }
       
        BigInteger r = g.modPow(k, p);
        BigInteger s = k.modInverse(p.subtract(BigInteger.ONE)).multiply(c.subtract((((ElGamalPrivateKeyParameters)key).getX()).multiply(r)).mod(p.subtract(BigInteger.ONE))).mod(p.subtract(BigInteger.ONE)); 
        
        BigInteger[]  sign = new BigInteger[2];

        sign[0] = r;
        sign[1] = s;

        return sign;
    }
    
     /**
     * Verifies the message against the signature values r and s.
     *
     * @param message the message that was supposed to have been signed.
     * @param r the r value of the signature.
     * @param s the s value of the signature.
     */
    public boolean verifySignature(CipherText cipher, BigInteger[] sign)
    {       
        BigInteger      c = cipher.getC();
        
        BigInteger      p = key.getParameters().getP();
        BigInteger      g = key.getParameters().getG();
                
        if(sign[0].compareTo(BigInteger.ZERO)==0 || sign[0].compareTo(p)==1)
        {
            return false;
        }
        
        if(sign[1].compareTo(BigInteger.ZERO)==0 || sign[1].compareTo(p.subtract(BigInteger.ONE))==1)
        {            
            return false;
        }
        
        BigInteger  u1 = g.modPow(c, p);
        BigInteger  u2 = ((ElGamalPublicKeyParameters)key).getY().modPow(sign[0], p).multiply(sign[0].modPow(sign[1], p)).mod(p);
        
        return u1.equals(u2);        
    }        
    
    
}
