package org.pnyka.cryptolib.ZKP;

import java.math.BigInteger;
import org.bouncycastle.crypto.params.ElGamalParameters;
import org.bouncycastle.crypto.params.ElGamalPrivateKeyParameters;
import org.bouncycastle.crypto.params.ElGamalPublicKeyParameters;

/**
 *
 *Discrete Logarithm equality ZKP.
 *
 * Suppose prover knows that two publically known quantities
 * <i>x = g<sup>l</sup></i> and <i>y = h<sup>l</sup></i> have the same discrete logarithms <i>l</i>
 * (to publically known respective bases g and h) in some group
 * of large prime order <i>P</i>. He wishes to convince verifier
 * of this  but without revealing what <i>l</i> is.
 *
 * The procedure is as follows:<ol>
 * <li>1. Prover chooses random r mod G (but keeps it private)</li>
 * <li>2. Prover computes and prints <i>a = g<sup>r</sup></i> and <i>b = h<sup>r</sup></i></li>
 * <li>3. Verifier chooses random <i>c mod G</i> and tells it to Prover</li>
 * <li>4. Prover computes and prints <i>z = r + lc mod G</i></li>
 * <li>5. Verifier verifies that <i>g<sup>z</sup> = ax<sup>c</sup></i> and <i>h<sup>z</sup> = by<sup>c</sup>.</i></li>
 * </ol>
 */
public class DisLogEqZKPAlg extends BaseZKP
{
    static BigInteger ONE  = BigInteger.valueOf(1);
    public BigInteger a, b, c, z;
    private boolean verification;
       
   /**
    * @param p parameter p
    * @param g parameter g
    * @param h parameter h
    */   
    public DisLogEqZKPAlg(BigInteger p, BigInteger g, BigInteger h)
    {
        super(p, g, h);
    }
    
    /**
     * Step 2 of ZKP. Prover computes and prints <i>a = g<sup>r</sup></i> and <i>b = h<sup>r</sup></i>.
     * @param r random value 
     */
    public void computeAB(BigInteger r)
    {
        a = g.modPow(r, p);
        b = h.modPow(r, p);
    }
    
     
    /**
     * Step 3 of ZKP. Verifier chooses random <i>c</i> mod <i>G</i> and tells it to Prover.
     */
    public void setChallenge()
    {
        c = BaseZKP.selectRandom(p);
    }
     
    /**
     * Step 4 of ZKP. Prover computes and prints <i>z = r + lc</i> mod <i>G</i>.
     * @param r the random r that prover chooses at the first step
     * @param l the logarithm <i>l</i>
     */
    public void computeZ(BigInteger r, BigInteger l)
    {
        z = r.add(l.multiply(c));        
    }
    
    /**
     * Step 5 of ZKP. Verifier verifies that <i>g<sup>z</sup> = ax<sup>c</sup></i> and <i>h<sup>z</sup> = by<sup>c</sup>.</i>.
     * @param x (x = g<sup>l</sup>)
     * @param y (y = h<sup>l</sup>) 
     * @return verification result
     */
    public boolean verify(BigInteger x,  BigInteger y)
    {
        BigInteger x1 = g.modPow(z, p); //g^z
        BigInteger x2 = a.multiply(x.modPow(c, p)).mod(p); 
        BigInteger y1 = h.modPow(z, p); //h^z
        BigInteger y2 = b.multiply(y.modPow(c, p)).mod(p);

                
        if(x1.equals(x2) && y1.equals(y2))
            verification = true;
        else 
            verification = false;
        
        return verification;
    }    
           
}
