package org.pnyka.cryptolib.crypto;

import java.math.BigInteger;
import java.security.SecureRandom;
import java.sql.Timestamp;
import org.bouncycastle.crypto.params.ElGamalPrivateKeyParameters;
import org.bouncycastle.crypto.params.ElGamalPublicKeyParameters;

/**
 * ElGamal encryption/decryption scheme.
 * 
 * Private and public keys are already produced and have the following format respectively:
 * Private Key: (p, g, k)
 * Public  Key: (p, g, h) 
 *
 * <b>Encryption</b> consists of the following steps:
 * 
 *  <ol>
 *      <li>Chose a random number <i>r </i>in the set{0,...., <i>q</i>-1}</li>
 *      <li>Compute value <i>z = g<sup>r</sup></i> mod <i>p</i></li>
 *      <li>Compute value <i> c = m h<sup>r</sup></i>mod </span><i>p</i>, <i>m </i>is the plaintext</span></li>
 *  </ol>
 *  <p>The pair (<i>z, c</i>)is the encrypted message </p>
 *
 * <b>Decryption</b> consists of the following step:
 * <ol>
 * Compute: <i>c z<sup>-K </sup></i>=<i> m * h<sup>y</sup> * g<sup>-ky&nbsp;</sup>=</i> <i>m</i> *<i> g<sup>ky</sup></i> * <i>g<sup>-ky</sup></i>= <i>m</i>
 * </ol>
 * 
 */

public class CipherText
{
    private Timestamp ts;
    private BigInteger z, c, r;
    
    /** Creates a new instance of ElGamalCipher */
    public CipherText(BigInteger z, BigInteger c)
    {
        this.z = z;
        this.c = c;
    }
    
    /** Creates a new instance of ElGamalCipher with the reencryption exponent r, if a message is reencrypted */
    public CipherText(BigInteger z, BigInteger c, BigInteger r)
    {
        this.z = z;
        this.c = c;
        this.r = r;
    }  
    
   /** Creates a new instance of ElGamalCipher with an SQL timestamp ts */
    public CipherText(BigInteger z, BigInteger c,  Timestamp ts)
    {
        this.z = z;
        this.c = c;
        this.ts = ts;
    }  
    
   /**
    * Encrypts a message using ElGamal encryption algorithm.
    * @param plain
    * @param pubKeyParams
    * @return ciphertext
    */
    public static CipherText elGamalEncrypt(PlainText plain, ElGamalPublicKeyParameters pubKeyParams)
    {
        SecureRandom sr = new SecureRandom();        
        BigInteger p = pubKeyParams.getParameters().getP();
        BigInteger g = pubKeyParams.getParameters().getG();
        BigInteger h = pubKeyParams.getY();
        
        int pBitLength = p.bitLength();
        BigInteger r = new BigInteger(pBitLength, sr);
                
        while (r.equals(BigInteger.ZERO) || (r.compareTo(p.subtract(BigInteger.valueOf(2))) > 0))
        {
            r = new BigInteger(pBitLength, sr);
        }            
        /////// (z,c) //////////
        BigInteger z = g.modPow(r, p);//z = g^r mod p
        BigInteger c = plain.getM().multiply(h.modPow(r, p)).mod(p);
                
        return new CipherText(z, c);
    }
   
    /**
    * Reencrypts a message using ElGamal encryption algorithm.
    * @param plain
    * @param pubKeyParams
    * @return ciphertext with r
    */
    public static  CipherText elGamalEncrypt(CipherText encM, ElGamalPublicKeyParameters pubKeyParams)
    {
        SecureRandom sr = new SecureRandom(); 
        BigInteger p = pubKeyParams.getParameters().getP();
        BigInteger g = pubKeyParams.getParameters().getG();
        BigInteger h = pubKeyParams.getY();
        
        int pBitLength = p.bitLength();
        BigInteger r = new BigInteger(pBitLength, sr);
        
        while (r.equals(BigInteger.ZERO) || (r.compareTo(p.subtract(BigInteger.valueOf(2))) > 0))
        {
            r = new BigInteger(pBitLength, sr);
        }            
              
       /////// (z',c') //////////
       BigInteger z = encM.getZ();
       BigInteger c = encM.getC();
       BigInteger zNew = z.multiply(g.modPow(r, p)).mod(p);//z' = g^r*g^q = g^s, s=r+q
       BigInteger cNew = c.multiply(h.modPow(r, p)).mod(p);//c' = h^r*h^q*M = h^s*M
     
       return new CipherText(zNew, cNew, r);
    }
    
   
   /**
    * Decrypts a message according to the ElGamal algorithm.
    * @param encMsg the encrypted message
    * @param params p, g values
    * @param privateKey the key for decryption
    * @return decrypted message
    */
   public static PlainText elGamalDecrypt(CipherText encM, ElGamalPrivateKeyParameters privKeyParams)
   {
       BigInteger a = privKeyParams.getX(); // private key
       BigInteger p = privKeyParams.getParameters().getP();
       BigInteger y1 = encM.getZ(); // cipher y1
       BigInteger y2 = encM.getC(); // cipher y2
       BigInteger x = null;
    
      x = y1.modPow(a, p).modInverse(p).multiply(y2).mod(p);
    
      return new PlainText(x);
   }
   
   /**
    * Decrypts partialy a message according to the ElGamal algorithm.
    * It is used when the key is shared between many keyholders 
    * @param encMsg the encrypted message
    * @param params p, g values
    * @param privateKey the key for decryption
    * @return partially decrypted message
    */
   public static CipherText elGamalDecryptPart(CipherText encM, ElGamalPrivateKeyParameters privKeyParams)
   {
        return new CipherText(encM.getZ(), elGamalDecrypt(encM, privKeyParams).getM());
   }
    
    
    /**
     * @return cipher z part
     */
    public BigInteger getZ()
    {
        return z;
    }

   /**
    * @return cipher c part
    */
    public BigInteger getC() 
    {
        return c;
    }
    
    /**
    * @return reencryption exponent r
    */
    public BigInteger getR() 
    {
        return r;
    }  
    
     /**
    * @return TimeStamp
    */
    public Timestamp getTimeStamp() 
    {
        return ts;
    }  
    
}