package org.pnyka.cryptolib.math;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.Date;

/**
 * Computation of Discrete Logarithm. Tow methods are provided: Shank's baby-step giant-step and the brute-force search.
 * 
 * <p>Baby-step giant-step method
 * Input: p, a, b
 * 
 * Output: x, where a<sup>x</sup> = b mod p</p>
 *
 * <ol>
 *  <li>Compute </font>&#945;<sup>mj </sup>mod p, 0 &#8804; j &#8804; m-1</li>
 *  <li>Sort m ordered pairs (j, &#945;<sup>mj</sup>mod p), obtaining list L1</li>
 *  <li>Compute &#946;<sup>&#945;-i</sup>mod p, 0 &#8804; i &#8804; m-1</li>
 *  <li>Sort m ordered pairs (i, &#946;<sup>&#945;-i</sup>mod p), obtaining list L2</li>
 *  <li>Find a pair (j,y) <font face="Symbol">?</font> L1 and a pair (i,y) <font face="Symbol">? </font>L2 (i.e., a pair having identical 2nd coordinates)</li>
 *  <li>Output mj +i mod(p-1).(&#945;<sup>mj</sup> =y= &#946;&#945;<sup>-i</sup>, &#945;<sup>mj+i</sup>=&#946; &#8594;log<sub>&#945;</sub>&#946;=mj+i)</li>
 *  </ol>
 *
 * Complexity : O(m) time, O(m) memory<br><br>
 *
 *
 *
 *  Brute-force method searches all exponents until the relationship a<sup>x</sup> = b mod p is satisfied.
 *  
 */
public class DiscreteLog
{
    /**
     * Computes the Discrete Logarithm using Shank's baby-step giant-step method.
     * @param a 
     * @param b
     * @param p
     * @return x the discrete logarithm
     */
     public static BigInteger babyGiantStep(BigInteger a, BigInteger b, BigInteger p)
    {
         BigInteger x = null;
         
          //Step 1: square root of p-1 (round_up)
          BigDecimal sqrt;
          BigSquareRoot app = new BigSquareRoot ();
          app.setTraceFlag (true);
          BigInteger n = p.subtract(BigInteger.valueOf(1));
                 
          int length = n.toString().length ();
          if (length > 20) 
          {
                app.setScale (length / 2);
          }
          BigInteger m = app.get(n).round(new MathContext(2, RoundingMode.CEILING)).toBigInteger();
        
          //step 2: compute a^m mod p
          BigInteger res = a.modPow(m, p);
         
       
       BigInteger brI = BigInteger.valueOf(0), brJ = BigInteger.valueOf(0);
       BigInteger j = BigInteger.ZERO;
       BigInteger i = BigInteger.ZERO;

       while(j.compareTo(p)<0)
       {
           BigInteger temp1 = a.modPow(m.multiply(j), p);
           
           while(i.compareTo(p)<0)
           {
                BigInteger temp2 = a.modPow(i, p).modInverse(p).multiply(b).mod(p); 
                if (temp1.compareTo(temp2) == 0)
                {                    
                    brI = i;
                    brJ = j;
                    break;
                }
               i = i.add(BigInteger.ONE);
           }
           
           j = j.add(BigInteger.ONE);
          
           if(brI.compareTo(BigInteger.valueOf(0))!=0 && brJ.compareTo(BigInteger.valueOf(0))!=0)
                break;
           
       }
       
       long stop_time = (new Date()).getTime();
       
       //Step 6: compute x = mj +i mod(p-1)
       x = m.multiply(brJ).add(brI.mod(p.subtract(BigInteger.valueOf(1))));
      
       return x;
    }
     
     
     
    /**
     * Computes the Discrete Logarithm with brute-force search.
     * @param a 
     * @param b
     * @param p
     * @return x the discrete logarithm
     */ 
     
    public static BigInteger bruteForce(BigInteger a, BigInteger b, BigInteger p)
    {
        BigInteger x = BigInteger.ZERO;
        
        while(a.modPow(x, p).compareTo(b.mod(p))!=0)
        {
            x = x.add(BigInteger.ONE);
        }
        
        return x;
    }
     
}
