package org.pnyka.cryptolib.prng;

import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Random;

/**
 * The RSA random number generator.
 *
 * <p>
 * The RSA is a "cryptographically secure" random number
 * generator.
 * </p>
 *
 * <p>
 * The Algorithm:
 * </p>
 * <ol>
 * <li>
 *  (setup) generate two secret prime numbers <i>p</i>, <i>q</i> such that
 *   <i>p</i> &ne; <i>q</i>, <i>p</i> &equiv; 3 mod 4, <i>q</i> &equiv; 3 mod 4.
 * </li>
 * <li> (setup) compute <i>n</i> = <i>pq</i>.  <i>n</i> can be re-used, but
 *   <i>p</i>, and <i>q</i> are secret and should be disposed of. Compute <i>phi =
 *    (p-1)(q-1)</i>. <br>Select a random integer <i>e</i>, 1 < <i>e</i> < <i>phi</i>,
 *    such that gcd(<i>e</i>, <i>phi</i>) = 1.
 * </li>
 * <li> Generate a (secure) random seed <i>s</i> in the range [1, <i>n</i> -1].
 * <li> Compute a single random bit with:
 *   <ol>
 *   <li> <i>x</i> = <i>x</i><sup>e</sup> mod <i>n</i></li>
 *   <li> return Least-Significant-Bit(<i>x</i>) (i.e. <i>x</i> & 1)</li>
 *   </ol>
 * Repeat as necessary.
 * </li>
 * </ol>
 *
 * <p>
 * More details are in  the <a href="http://www.cacr.math.uwaterloo.ca/hac/"><i>Handbook of Applied Cryptography</i></a>,
 * <a href="http://www.cacr.math.uwaterloo.ca/hac/about/chap5.pdf">Section 5.5.1</a>
 * </p>
 *
 */

public class RSARandomGenerator implements RandomGenerator
{    
    // pre-compute a few values
    private static final BigInteger two = BigInteger.valueOf(2L);
    private static final BigInteger three = BigInteger.valueOf(3L);
    private static final BigInteger four = BigInteger.valueOf(4L);
       
    private BigInteger n;
    private BigInteger e;
    private BigInteger state;
    
    
    public RSARandomGenerator(int bits) //Constructor 1
    {
	this(bits, new Random());
    }
     
    public RSARandomGenerator(int bits, Random rand) //Constructor 2
    {
	this(generateNE(bits, rand));     
    }
    
    public RSARandomGenerator(BigInteger[] NandE) //Constructor 3
    {
	this(NandE[0], NandE[1], SecureRandom.getSeed(NandE[0].bitLength() / 8));        
    }
    
   
    /**
     * A constructor to specify both the n-value and the seed to the
     * Blum-Blum-Shub algorithm.
     *
     * @param n The n-value using a BigInteger
     * @param e The e-value using a BigInteger
     * @param seed The seed value using a byte[] array.
     */
    public RSARandomGenerator(BigInteger n, BigInteger e, byte[] seed) //Constructor 4
    {
	this.n = n;
	this.e = e;
        setSeed(seed);
    }        
   

    /**
     * Generate appropriate prime number for use in RSA prng.
     *
     * This generates the appropriate primes needed to compute the
     * "n-value" for RSA.
     *
     * @param bits Number of bits in prime
     * @param rand A source of randomness
     */
    private static BigInteger getPrime(int bits, Random rand) 
    {
	BigInteger p;
	while (true) 
        {
	    p = new BigInteger(bits, 100, rand);
	    if (p.mod(four).equals(three)) //p mod 4 = 3
		break;
	}
	return p;
    }

    /**
     * This generates the "n value" -- the multiplication of two equally sized
     * random prime numbers -- for use in the RSA algorithm.
     *
     * @param bits The number of bits of security
     * @param rand A random instance to aid in generating primes
     * @return A BigInteger, the <i>n</i>.
     */
    public static BigInteger[] generateNE(int bits, Random rand) 
    {
	BigInteger p = getPrime(bits/2, rand);
	BigInteger q = getPrime(bits/2, rand);

	// make sure p != q (almost always true, but just in case, check)
	while (p.equals(q)) 
        {
	    q = getPrime(bits, rand);
	}
        
        BigInteger n = p.multiply(q);
        
        BigInteger phi = p.subtract(BigInteger.ONE).multiply(q.subtract(BigInteger.ONE));
        
        BigInteger e = generateE(bits, rand, phi);

        BigInteger[] NandEValues = new BigInteger[2];
        NandEValues[0] = n;
        NandEValues[1] = e;
        
	return NandEValues;
    }
    
    
     //select random integer in the interval 1< e < phi such that gcd(e, phi) = 1
     private static BigInteger generateE(int bits, Random rand, BigInteger phi)
     {        
         BigInteger epsilon = new BigInteger(bits, rand);
                         
         while(epsilon.gcd(phi).compareTo(BigInteger.ONE) != 0)
         {
              epsilon = new BigInteger(bits, rand);
         }
         
         return epsilon;
     }

    
    /**
     * Sets or resets the seed value and internal state
     *
     * @param seedBytes The new seed.
     */
    public void setSeed(byte[] seedBytes)
    {
	BigInteger seed = new BigInteger(1, seedBytes);
	state = seed.mod(n);       
    }

    /**
     * Returns a random number 
     *
     * See also java.util.Random#next
     * @param numBits the length in bits of random number
     * @return a random Biginteger
     */
    public BigInteger nextNumber(int numBits)
    {
	BigInteger result = BigInteger.ZERO;
        int result2 = 0;
        
	for (int i = numBits; i != 0; --i) 
        {
	    state = state.modPow(e, n);
            result = result.shiftLeft(1).or((state.testBit(0) == true ? BigInteger.ONE : BigInteger.ZERO));
	    result2 = (result2 << 1) | (state.testBit(0) == true ? 1 : 0);
	}        
           
	return result;
    }

          
}
