package org.pnyka.timestamper;

import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.cert.CertStore;
import java.security.cert.CertStoreException;
import java.security.cert.CertStoreParameters;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CollectionCertStoreParameters;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Random;
import javax.security.auth.x500.X500Principal;

import org.bouncycastle.asn1.ASN1OctetString;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.x509.ExtendedKeyUsage;
import org.bouncycastle.asn1.x509.KeyPurposeId;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.tsp.TSPAlgorithms;
import org.bouncycastle.tsp.TSPException;
import org.bouncycastle.tsp.TimeStampRequest;
import org.bouncycastle.tsp.TimeStampRequestGenerator;
import org.bouncycastle.tsp.TimeStampResponse;
import org.bouncycastle.tsp.TimeStampResponseGenerator;
import org.bouncycastle.tsp.TimeStampTokenGenerator;
import org.bouncycastle.x509.X509V3CertificateGenerator;

public class TSProtocol
{    
   /**
    * Generates a TimeStamp Request.
    * @param md digest of the message that will be timestamp 
    * @param tspAlg algorithm  
    * @param nonce if true, a random number must be included to request
    * @param certReq if true, the TSA's public key certificate must be provided by the TSA in the certificates field from the SignedData structure in that response.
    * @return request
    */
    public TimeStampRequest tspRequest(byte[] md, String tspAlg, boolean nonce, boolean certReq)
    {
       TimeStampRequest request = null; 
       /* The messageImprint field SHOULD contain the hash of the datum to be 
       time-stamped.  The hash is represented as an OCTET STRING.  Its 
       length MUST match the length of the hash value for that algorithm 
       (e.g., 20 bytes for SHA-1 or 16 bytes for MD5). */
       ASN1OctetString octet =	new DEROctetString(md);  
       TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
           
       reqGen.setCertReq(certReq);               
       
       if(nonce == true)
       {
           if( tspAlg == "MD5" )
               request = reqGen.generate(TSPAlgorithms.MD5, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "GOST3411" )
               request = reqGen.generate(TSPAlgorithms.GOST3411, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "RIPEMD128" )
               request = reqGen.generate(TSPAlgorithms.RIPEMD128, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "RIPEMD160" )
               request = reqGen.generate(TSPAlgorithms.RIPEMD160, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "RIPEMD256" )
               request = reqGen.generate(TSPAlgorithms.RIPEMD256, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "SHA1" )
               request = reqGen.generate(TSPAlgorithms.SHA1, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "SHA224" )
               request = reqGen.generate(TSPAlgorithms.SHA224, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "SHA256" )
               request = reqGen.generate(TSPAlgorithms.SHA256, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "SHA384 " )
               request = reqGen.generate(TSPAlgorithms.SHA384, octet.getOctets(), new BigInteger(16, new Random()));
           else if( tspAlg == "SHA512" )
               request = reqGen.generate(TSPAlgorithms.SHA512, octet.getOctets(), new BigInteger(16, new Random()));
                            
       }
       else 
       {
           if( tspAlg == "MD5" )
               request = reqGen.generate(TSPAlgorithms.MD5, octet.getOctets());
           else if( tspAlg == "GOST3411" )
               request = reqGen.generate(TSPAlgorithms.GOST3411, octet.getOctets());
           else if( tspAlg == "RIPEMD128" )
               request = reqGen.generate(TSPAlgorithms.RIPEMD128, octet.getOctets());
           else if( tspAlg == "RIPEMD160" )
               request = reqGen.generate(TSPAlgorithms.RIPEMD160, octet.getOctets());
           else if( tspAlg == "RIPEMD256" )
               request = reqGen.generate(TSPAlgorithms.RIPEMD256, octet.getOctets());
           else if( tspAlg == "SHA1" )
               request = reqGen.generate(TSPAlgorithms.SHA1, octet.getOctets());
           else if( tspAlg == "SHA224" )
               request = reqGen.generate(TSPAlgorithms.SHA224, octet.getOctets());
           else if( tspAlg == "SHA256" )
               request = reqGen.generate(TSPAlgorithms.SHA256, octet.getOctets());
           else if( tspAlg == "SHA384 " )
               request = reqGen.generate(TSPAlgorithms.SHA384, octet.getOctets());
           else if( tspAlg == "SHA512" )
               request = reqGen.generate(TSPAlgorithms.SHA512, octet.getOctets());
       }

       return request;
    }
    
  /**
   * Generates a TimeStamp Response
   * @param req the request that was sent to TSA
   * @param key CA's private key
   * @param cert TSA's certificate
   * @param ordering if true, every time-stamp token from the same TSA can always be ordered based on the dateAndTime field
   * @param dateAndTime date and time the timestamp was created, expressed as UTC time
   * @return response
   */ 
   public TimeStampResponse tspResponse( TimeStampRequest req, PrivateKey key, X509Certificate cert, boolean ordering, long dateAndTime, String tspAlg )
          throws TSPException, InvalidAlgorithmParameterException, CertStoreException, NoSuchAlgorithmException, NoSuchProviderException
   {
       TimeStampResponse response = null;
       TimeStampTokenGenerator tokenGen = null;
       
       if(tspAlg == "MD5")
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.MD5, "1.2");
       else if( tspAlg == "GOST3411" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.GOST3411, "1.2");
       else if( tspAlg == "RIPEMD128" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.RIPEMD128, "1.2");
       else if( tspAlg == "RIPEMD160" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.RIPEMD160, "1.2");
       else if( tspAlg == "RIPEMD256" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.RIPEMD256, "1.2");
       else if( tspAlg == "SHA1" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.SHA1, "1.2");
       else if( tspAlg == "SHA224" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.SHA224, "1.2");
       else if( tspAlg == "SHA256" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.RIPEMD256, "1.2");
       else if( tspAlg == "SHA384 " )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.SHA384, "1.2");
       else if( tspAlg == "SHA512" )
           tokenGen = new TimeStampTokenGenerator(key, cert, TSPAlgorithms.SHA512, "1.2");
                  
       tokenGen.setOrdering(ordering);
       CertStoreParameters csp = new CollectionCertStoreParameters();
       /* CertStore object: periexei ta certificates kai ta CRLs pou tha symperilhfthoun
       sto CMSSignedData tou token pou tha epistrepsei h TSA ston voter*/
       CertStore certStore = CertStore.getInstance("Collection", csp); 
       tokenGen.setCertificatesAndCRLs(certStore);
       //TimeStampToken token = tokenGen.generate(req, BigInteger.valueOf(234), new Date(), "BC" );
       TimeStampResponseGenerator respGen = new TimeStampResponseGenerator(tokenGen, TSPAlgorithms.ALLOWED);
       response = respGen.generate(req, BigInteger.valueOf(1234), new Date(dateAndTime), "BC");
       //System.out.println("date created "+DateFormat.getDateTimeInstance().format(new Date(dateAndTime))); 
            
       return response;
    }
   
   /**
    * Generates a X509 Certificates.
    * @param keypairEA TSA's private and public key
    * @param startDate time from which certificate is valid  
    * @param expiryDate time after which certificate is not valid
    * @param signAlg identifies the algorithm used by the CA to sign the certificate
    * @param serialNumber certificate's serial number
    * @param subName name of the entity whose public key the certificate identifies
    * @param issuName name of the entity that signed the certificate
    * @param provider cryptography provider name
    * @return cert
    */
   public X509Certificate Certificate(KeyPair keypairCA, KeyPair keypairTSA, Date startDate, Date expiryDate, 
          String signAlg, BigInteger serialNumber, String subName, String issuName, String provider) 
          throws CertificateEncodingException, NoSuchProviderException, NoSuchAlgorithmException, SignatureException, InvalidKeyException
   {
        X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
        X500Principal subjectName = new X500Principal(subName);//
        
        //Stoixeia EA kai certificate
        certGen.setSerialNumber(serialNumber);
        certGen.setIssuerDN(new X509Name(issuName));//Issuer Distinguished Name: CA's name
        certGen.setNotBefore(startDate);
        certGen.setNotAfter(expiryDate);
        certGen.setSubjectDN(subjectName);//Subject Distinguished Name: EA's name
        certGen.setSignatureAlgorithm(signAlg);
        certGen.setPublicKey(keypairTSA.getPublic());
        certGen.addExtension(X509Extensions.ExtendedKeyUsage, true, new ExtendedKeyUsage(KeyPurposeId.id_kp_timeStamping));
       
        X509Certificate cert = certGen.generate(keypairCA.getPrivate(), provider); 
            
        return cert;
    }
    
   /*public static void main(String[] args) throws IOException, NoSuchAlgorithmException, ClassNotFoundException, java.security.cert.CertificateException, NoSuchProviderException, InstantiationException, NoSuchMethodException, TSPException, IllegalAccessException, InvocationTargetException, SignatureException, InvalidKeyException, InvalidAlgorithmParameterException, CertStoreException, CMSException 
    {     
        TSProtocol tsProto = new TSProtocol();
        Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
               
        //public kai private keys ths TSA 
        KeyPairGenerator keyGenTSA = KeyPairGenerator.getInstance("RSA");
        keyGenTSA.initialize(512);
        KeyPair keypairTSA = keyGenTSA.genKeyPair();
        
        //keypair of the certifying authority 
        KeyPairGenerator keyGenCA = KeyPairGenerator.getInstance("RSA");
        keyGenCA.initialize(512);
        KeyPair keypairCA = keyGenCA.genKeyPair();
                
        //TSA's certificate creation
        Date startDate = new Date(1154371345647L);            // time from which certificate is valid
        Date expiryDate = new Date(1194371345647L);           // time after which certificate is not valid
        String signAlg = "MD5withRSA";
        BigInteger SN = BigInteger.valueOf(1234);             // certificate's serial number 
        String subName = "CN=Election Authority 1, OU=Authority, O=Gov, C=GR";
        String issuName = "CN=Certificate Authority XXX";
        String provider = "BC";
        X509Certificate cert = tsProto.Certificate(keypairCA, keypairTSA, startDate, expiryDate, signAlg, SN, subName, issuName, provider);
        CertStoreParameters csp = new CollectionCertStoreParameters();
        CertStore certStore = CertStore.getInstance("Collection", csp); 
        //TSPUtil valid = new TSPUtil();
        //valid.validateCertificate(cert);//certificate validation
           
        //---- h EA sxhmatizei to request -----
        byte[] md = MessageDigest.HashValue("TSP".getBytes());//to original mnm pou thelei na kanei timestamping
        TimeStampRequest req =  tsProto.tspRequest(md, "MD5", true, false);//o voter stelnei request gia Timestamp
        
        //----- h TSA stelnei response ---        
        TimeStampResponse response = tsProto.tspResponse(req, keypairTSA.getPrivate(), cert, false, 1174371345647L, "MD5"); //response me token apo TSA
        int status  = response.getStatus();
        PKIFailureInfo fail = response.getFailInfo();
         
        //------ o voter dexetai response------
        TimeStampToken tok = response.getTimeStampToken(); //token
        tok.validate(cert, "BC"); 
        
        //////////////////
        byte[] reqEnq = req.getEncoded();//h EA dhmiourgei thn ASN1 encodable ekdosh tou request pou sxhmatise kai prepei na steilei sthn TSA
        TimeStampRequest receivedReq = new TimeStampRequest(reqEnq);//otan lavei h TSA to hn ASN1 encodable ekdosh tou request, ksanasxhmatizei to request
        
                
        
        System.out.println("date created "+DateFormat.getDateTimeInstance().format(startDate));
        System.out.println("date created "+DateFormat.getDateTimeInstance().format(expiryDate));
        TSPUtil.validateCertificate(cert);
        TimeStampTokenInfo tstInfo = tok.getTimeStampInfo();
        System.out.println("date created "+DateFormat.getDateTimeInstance().format(tstInfo.getGenTime()));
        CMSSignedData sd = tok.toCMSSignedData();      
        byte[] en = sd.getEncoded();
             
      }*/
        
  }
